# -*- coding: utf-8 -*-
import logging

import click

from chanjo.store import ChanjoAPI
from chanjo.store.utils import filter_samples
from .utils import dump_json

logger = logging.getLogger(__name__)


@click.group()
@click.pass_context
def calculate(context):
    """Calculate simple metrics from the database."""
    context.api = ChanjoAPI(uri=context.obj['database'])


@calculate.command()
@click.argument('samples', nargs=-1)
@click.pass_context
def mean(context, samples):
    """Report mean coverage for a list of samples."""
    api = context.parent.api
    query = filter_samples(api.query(), sample_ids=samples)
    results = ({'sample_id': sample_id, 'metrics': data}
               for sample_id, data in api.means(query))
    dump_json(*results)


@calculate.command()
@click.argument('gene_ids', nargs=-1)
@click.pass_context
def gene(context, gene_ids):
    """Report aggregate statistics for particular genes."""
    results = context.parent.api.gene(*gene_ids)
    dump_json(*results)


@calculate.command()
@click.option('-s', '--sample', help='limit to a single sample')
@click.option('-p', '--per', type=click.Choice(['exon', 'sample']),
              help='report stats per sample/exon')
@click.argument('chromosome', type=str)
@click.argument('start', type=int, required=False)
@click.argument('end', type=int, required=False)
@click.pass_context
def region(context, sample, per, chromosome, start, end):
    """Report mean statistics for a region of exons."""
    api = context.parent.api
    if start is None:
        logger.debug('region id detected, parse string')
        results = api.region_alt(chromosome, sample_id=sample, per=per)
    else:
        query = api.region(chromosome, start, end, sample_id=sample, per=per)
        results = ({'exon_id': exon_id, 'metrics': data}
                   for exon_id, data in query)
    if per == 'exon':
        dump_json(*results)
    else:
        dump_json(results)



# -*- coding: utf-8 -*-
from .models import Exon, ExonStatistic, Gene, Sample, Transcript


class ChanjoConverterMixin(object):

    """Mixin to provide conversions between genomic elements."""

    def gene_transcripts(self, *gene_ids):
        """Fetch transcripts related to a list of genes.

        Args:
            *gene_ids (List[str]): gene ids

        Returns:
            List[Transcript]: transcript objects related to the genes
        """
        query = (self.query(Transcript)
                     .distinct(Transcript.transcript_id)
                     .join(Exon.transcripts)
                     .filter(Gene.gene_id.in_(gene_ids)))
        return query

    def gene_exons(self, *gene_ids):
        """Fetch exons related to a list of genes.

        Args:
            *gene_ids (List[str]): gene ids

        Returns:
            List[Exon]: exon objects related to the genes
        """
        query = (self.query(Exon)
                     .distinct(Exon.exon_id)
                     .join(Exon.transcripts, Transcript.gene)
                     .filter(Gene.gene_id.in_(gene_ids)))
        return query

    def transcript_genes(self, transcript_ids, db_ids=False):
        """Fetch a lists of genes related to some exons.

        Args:
            transcript_ids (List[str]): transcript ids to convert to genes
            db_ids (Optional[bool]): if sending in primary key ids

        Returns:
            List[Gene]: gene models related to the transcripts
        """
        results = self.query(Gene).join(Gene.transcripts)
        if db_ids:
            results = results.filter(Transcript.id.in_(transcript_ids))
        else:
            condition = Transcript.transcript_id.in_(transcript_ids)
            results = results.filter(condition)
        return results

    def transcript_exons(self, *transcript_ids):
        """Fetch a unique list of exons related to some transcripts.

        Args:
            transcript_ids (List[str]): transcript ids to look up

        Returns:
            List[tuple]: sample, metric, weighted average value
        """
        results = (self.query(Sample.sample_id, ExonStatistic.metric,
                              self.weighted_average)
                       .join(ExonStatistic.sample, ExonStatistic.exon,
                             Exon.transcripts)
                       .filter(Transcript.transcript_id.in_(transcript_ids))
                       .group_by(Sample.sample_id, ExonStatistic.metric))
        return results

    def exon_transcripts(self, exons_ids, db_ids=False):
        """Fetch a unique list of transcripts related to some exons.

        Args:
            exon_ids (List[str]): list of exon ids
            db_ids (Optional[bool]): if sending in primary key ids

        Returns:
            List[Transcript]: transcripts related to the exons
        """
        results = (self.query(Transcript)
                       .distinct(Transcript.transcript_id)
                       .join(Transcript.exons))
        if db_ids:
            results = results.filter(Exon.id.in_(exons_ids))
        else:
            results = results.filter(Exon.exon_id.in_(exons_ids))
        return results

    def exon_genes(self, *exon_ids):
        """Fetch unique genes related to a list of exons.

        Args:
            *exon_ids (List[str]): exon ids

        Returns:
            List[Gene]: gene objects related to the exons
        """
        query = (self.query(Gene)
                     .distinct(Gene.gene_id)
                     .join(Gene.transcripts, Transcript.exons)
                     .filter(Exon.exon_id.in_(exon_ids)))
        return query



# -*- coding: utf-8 -*-
from chanjo.store import Exon, Exon_Transcript


def test_Exon_Transcript():
    # test table name
    assert Exon_Transcript.name == 'exon__transcript'


def test_Interval():
    # set up
    exon = Exon(exon_id='exon1', chromosome='chr1', start=10, end=1000)

    assert len(exon) == 991



# -*- coding: utf-8 -*-
import logging
import urllib
import zipfile

from path import path

from .constants import BED_URL, DATABASE_URL

logger = logging.getLogger(__name__)


def pull(target_dir, force=False):
    """Download precompiled resources into a folder."""
    logger.debug('ensure target directory exists')
    target_path = path(target_dir)
    target_path.makedirs_p()

    database_path = target_path.joinpath('coverage.sqlite3')
    bed_path = target_path.joinpath('ccds.bed.zip')

    files = ((database_path, DATABASE_URL), (bed_path, BED_URL))
    for file_path, url in files:
        if path(file_path).exists() and not force:
            logger.warn('file already exists, skipping: %s', file_path)
        else:
            logger.info('downloading... [%s]', url)
            urllib.urlretrieve(url, file_path)

    logger.info('extracting BED file...')
    zip_ref = zipfile.ZipFile(bed_path, 'r')
    zip_ref.extractall(target_dir)

    logger.info('removing BED archive...')
    bed_path.remove_p()



import subprocess
import logging

from subprocess import CalledProcessError

from chanjo.log import get_log_stream


def run_sambamba(bam_file, region_file, outfile=None, cov_treshold=()):
    """Run sambamba from chanjo."""
    logger = logging.getLogger(__name__)
    logger = logging.getLogger("chanjo.sambamba")
    log_stream = get_log_stream(logger)
    sambamba_call = [
        "sambamba",
        "depth",
        "region",
        "--regions",
        region_file,
        bam_file
    ]

    if outfile:
        sambamba_call += ["-o", outfile]

    for coverage_treshold in cov_treshold:
        sambamba_call += ['-T', str(coverage_treshold)]

    logger.info("Running sambamba with call: %s", ' '.join(sambamba_call))
    try:
        subprocess.check_call(
            sambamba_call,
            stderr=log_stream,
        )
    except OSError as e:
        logger.critical("sambamba seems to not exist on your system.")
        raise e
    except CalledProcessError as e:
        logger.critical("Something went wrong when running sambamba. "
                        "Please see sambamba error output.")
        raise e

    logger.debug("sambamba run successfull")
    return



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from chanjo.store import Store
from chanjo.load import sambamba as load_sambamba
from chanjo.parse import sambamba as parse_sambamba
from get_bedrows import get_bedlines

bed_lines = get_bedlines()


class TestSambamba:
    def setup(self):
        self.store = Store('sqlite://')
        self.store.set_up()
        self.row_data = parse_sambamba.depth_output(bed_lines)

    def teardown(self):
        self.store.tear_down()
        self.store.session.close()

    def test_connection(self):
        self.store.dialect == 'sqlite'

    def test_rows(self):
        stats = load_sambamba.rows(self.store.session, self.row_data)
        all_stats = [stat for stat in stats]
        assert len(all_stats) == 4



def get_bedlines():
    bed_lines = ["# chrom\tchromStart\tchromEnd\tF3\tF4\tF5\tF6\tF7\t"
                    "readCount\tmeanCoverage\tpercentage10\tpercentage20"
                    "\tpercentage100\tsampleName",
                 "1\t69089\t70007\t1-69090-70007\t0\t+\tCCDS30547.1\tOR4F5\t"
                    "232\t25.4946\t57.9521\t36.0566\t5.55556\tADM992A10\t"]
    return bed_lines



# -*- coding: utf-8 -*-
import sys
import click
import logging

from chanjo.annotate.sambamba import run_sambamba


@click.command()
@click.argument('bam_file',
                    type=click.Path(exists=True),
)
@click.option('-e', '--exon_bed',
                type=click.Path(exists=True),
                help="Path to a bed file with exon coordinates"
)
@click.option('-g', '--gene_bed',
                type=click.Path(exists=True),
                help="Path to a bed file with gene coordinates"
)
@click.option('-t', '--cov_treshold',
                multiple=True,
                type=int,
                help="multiple thresholds can be provided,"\
                     "for each one an extra column will be added,"\
                     "the percentage of bases in the region"\
                     "where coverage is more than this value"
)
@click.option('-o', '--outfile',
                    type=click.Path(exists=False),
                    help='Specify the path to a file where results should be stored.'
)
@click.pass_context
def sambamba(context, bam_file, exon_bed, gene_bed, cov_treshold, outfile):
    """Run Sambamba from chanjo."""
    logger = logging.getLogger(__name__)
    # For testing only:
    logger = logging.getLogger("chanjo.cli.sambamba")
    logger.info("Running chanjo sambamba")

    if not (exon_bed or gene_bed):
        logger.warning("Please provide a region file in BED format")
        sys.exit()
    if exon_bed and gene_bed:
        logger.warning("Only one region file at a time")
        sys.exit()

    region_file = exon_bed
    if gene_bed:
        region_file = gene_bed

    try:
        run_sambamba(bam_file, region_file, outfile, cov_treshold)
    except Exception as e:
        logger.debug(e)
        click.Abort()



# -*- coding: utf-8 -*-
from chanjo.load.utils import exon, _exon_kwargs


DATA = {'chrom': 'chr1', 'chromStart': 100, 'chromEnd': 220, 'name': 'exon1',
        'score': 0, 'strand': '+', 'sampleName': 'sample1', 'readCount': 10,
        'meanCoverage': 6.341, 'thresholds': {10: 95.421, 20: 86.21, 100: 10.21}}


def test_exon():
    exon_obj = exon(DATA)
    assert exon_obj.chromosome == 'chr1'
    assert exon_obj.exon_id == 'exon1'


def test__exon_kwargs():
    kwargs = _exon_kwargs(DATA)
    assert kwargs['chromosome'] == 'chr1'
    assert kwargs['exon_id'] == 'exon1'



# -*- coding: utf-8 -*-
from toolz import cons

from chanjo.compat import iteritems
from chanjo.store import ExonStatistic, Sample

from .utils import get_or_build_exon


def rows(session, row_data, sample_id=None, group_id=None):
    """Handle rows of sambamba output.

    N.B. only handles single sample annotations.

    Args:
        session (Session): database session object
        row_data (dict): parsed sambamba output rows
        sample_id (Optional[str]): id to reference sample
        group_id (Optional[str]): id to group samples together

    Yields:
        ExonStatistic: stats model linked to exon and sample
    """
    if sample_id is None:
        # use first row to get information on sample
        first_row = next(iter(row_data))
        sample_id = first_row['sampleName']
        # place the first row back in the stream
        all_data = cons(first_row, row_data)
    else:
        all_data = row_data

    sample_obj = Sample(sample_id=sample_id, group_id=group_id)
    nested_stats = (row(session, data, sample_obj) for data in all_data)
    # flatten 2D nested list
    return (stat for stats in nested_stats for stat in stats)


def row(session, data, sample_obj):
    """Handle sambamba output row.

    Args:
        session (Session): database session object
        data (dict): parsed sambamba output row
        sample_obj (Sample): linked sample model

    Returns:
        List[ExonStatistic]: stats models linked to exon and sample
    """
    exon_obj = get_or_build_exon(session, data)
    stats = statistics(data, sample_obj, exon_obj)
    return stats


def statistics(data, sample_obj, exon_obj):
    """Create models from a sambamba output row.

    Args:
        data (dict): parsed sambamba output row
        sample_obj (Sample): linked sample model
        exon_obj (Exon): linked exon model

    Returns:
        List[ExonStatistic]: stats models linked to exon and sample
    """
    relationships = dict(sample=sample_obj, exon=exon_obj)
    stats = [ExonStatistic(metric='mean_coverage', value=data['meanCoverage'],
                           **relationships)]
    for threshold, value in iteritems(data['thresholds']):
        metric = "completeness_{}".format(threshold)
        stat = ExonStatistic(metric=metric, value=value, **relationships)
        stats.append(stat)

    return stats



# -*- coding: utf-8 -*-
import logging

from .models import Sample

logger = logging.getLogger(__name__)


def filter_samples(query, group_id=None, sample_ids=None):
    """Filter a query to a subset of samples.

    Will return an unaltered query if none of the optional parameters
    are set. `group_id` takes precedence over `sample_ids`.

    Args:
        query (Query): SQLAlchemy query object
        group_id (Optional[str]): sample group identifier
        sample_ids (Optional[List[str]]): sample ids

    Returns:
        Query: filtered query object
    """
    if group_id:
        logger.debug('filter based on group')
        return query.filter(Sample.group_id == group_id)
    elif sample_ids:
        logger.debug('filter based on list of samples')
        return query.filter(Sample.sample_id.in_(sample_ids))
    else:
        return query


def group_by_field(results):
    """Group items based on the initial field.

    Assumes a sorted list of results.

    Args:
        results (List[tuple]): list of fields to group

    Yields:
        str, dict: next group of results along with the group id
    """
    results = iter(results)
    current_group_id, metric, value = next(results)
    group = {metric: value}
    for group_id, metric, value in results:
        if current_group_id != group_id:
            yield current_group_id, group
            # reset the group
            current_group_id = group_id
            group = {}
        # store the metric under the correct group
        group[metric] = value
    yield current_group_id, group


def predict_gender(x_coverage, y_coverage):
    """Make a simple yet accurate prediction of a samples gender.

    The calculation is based on the average coverage across the X and Y
    chromosomes. Note that an extrapolation from a subsection of bases is
    usually quite sufficient.

    Args:
        x_coverage (float): estimated average coverage on X chromosome
        y_coverage (float): estimated average coverage on Y chromosome

    Returns:
        str: prediction, ['male', 'female', 'unknown']
    """
    # algoritm doesn't work if coverage is missing for X chromosome
    if x_coverage == 0:
        return 'unknown'
    # this is the entire prediction, it's usually very obvious
    elif (y_coverage > 0) and (x_coverage / y_coverage < 10):
        return 'male'
    else:
        # the few reads mapping to the Y chromosomes are artifacts
        return 'female'



# -*- coding: utf-8 -*-
from chanjo.config import remove_ansi


def test_remove_ansi():
    # Test a general case
    ansi_string = '[\x1b[32m?\x1b[0m] name: '
    assert remove_ansi(ansi_string) == '[?] name: '



# -*- coding: utf-8 -*-
from functools import partial
import re

from toolz import curry, get

# regular expression to delete ANSI escape sequences (UNIX color codes)
# ref: http://stackoverflow.com/questions/14693701
ansi_escape = re.compile(r'\x1b[^m]*m')

remove_ansi = partial(ansi_escape.sub, '')
remove_ansi.__doc__ = """Remove ANSI escape characters from a string.

.. code-block:: python

    >>> ansi_string = '[\x1b[32m?\x1b[0m] name: '
    >>> remove_ansi(ansi_string)
    '[?] name: '

Args:
    ansi_string (str): string, possibly containing ANSI escape characters

Returns:
    str: input string *guaranteed* without ANSI escape characters
"""


@curry
def rget(sequence, key, default=None):
    """Get element in a sequence or dict.

    Like toolz.get but with parameters in reverse order.

    Args:
        sequence (sequence or dict): sequence or dict
        key (str or int): key to access in sequence

    Returns:
        object: value behind the key
    """
    return get(key, sequence, default=default)



# -*- coding: utf-8 -*-
from chanjo.parse import sambamba

BASE_HEADERS = ['chrom', 'chromStart', 'chromEnd']
THRESHOLDS = ['percentage10', 'percentage20', 'percentage100']
EXTRA_HEADERS = ['F3', 'F4', 'F5']
SAMBAMBA_HEADERS = ['readCount', 'meanCoverage']

BASE_COLS = ['1', '69089', '70007']
EXTRA_COLS = ['1-69090-70007', '0', '+', 'CCDS30547.1', 'OR4F5']
SAMBAMBA_COLS = ['232', '25.4946']
THRESHOLD_COLS = ['57.9521', '36.0566', '5.55556']
SAMBAMBA_END_COLS = ['ADM992A10', '']


def test_expand_header_no_thresholds():
    header_row = BASE_HEADERS + SAMBAMBA_HEADERS + ['sampleName']
    header = sambamba.expand_header(header_row)
    assert header['readCount'] == 3
    assert header['thresholds'] == {}
    assert header['extraFields'] == slice(3, 3)
    assert header['sampleName'] == 5


def test_expand_header_single_threshold():
    header_row = (BASE_HEADERS + SAMBAMBA_HEADERS + THRESHOLDS[:1]
                  + ['sampleName'])
    header = sambamba.expand_header(header_row)
    assert header['readCount'] == 3
    assert header['thresholds'] == {10: 5}
    assert header['sampleName'] == 6


def test_expand_header_multi_thresholds():
    header_row = (BASE_HEADERS + SAMBAMBA_HEADERS + THRESHOLDS
                  + ['sampleName'])
    header = sambamba.expand_header(header_row)
    assert header['readCount'] == 3
    assert header['thresholds'] == {10: 5, 20: 6, 100: 7}
    assert header['sampleName'] == 8


def test_expand_header_extra_cols():
    header_row = (BASE_HEADERS + EXTRA_HEADERS + SAMBAMBA_HEADERS
                  + ['sampleName'])
    header = sambamba.expand_header(header_row)
    assert header['readCount'] == 6
    assert header['extraFields'] == slice(3, 6)
    assert header['sampleName'] == 8


def test_expand_row_basic():
    header = {'extraFields': slice(3, 3), 'readCount': 3, 'meanCoverage': 4,
              'thresholds': {}, 'sampleName': 5}
    row = BASE_COLS + SAMBAMBA_COLS + SAMBAMBA_END_COLS
    row_data = sambamba.expand_row(header, row)
    assert row_data['chrom'] == '1'
    assert row_data['chromStart'] == 69089
    assert row_data['chromEnd'] == 70007
    assert row_data['extraFields'] == []
    assert row_data['readCount'] == 232
    assert row_data['meanCoverage'] == 25.4946
    assert row_data['thresholds'] == {}
    assert row_data['sampleName'] == 'ADM992A10'


def test_expand_row_with_threshold():
    header = {'extraFields': slice(3, 3), 'readCount': 3, 'meanCoverage': 4,
              'thresholds': {10: 5, 20: 6, 100: 7}, 'sampleName': 8}
    row = BASE_COLS + SAMBAMBA_COLS + THRESHOLD_COLS + SAMBAMBA_END_COLS
    row_data = sambamba.expand_row(header, row)
    assert row_data['thresholds'][10] == 57.9521
    assert row_data['thresholds'][20] == 36.0566
    assert row_data['thresholds'][100] == 5.55556
    assert row_data['sampleName'] == 'ADM992A10'


def test_expand_row_with_extra_fields():
    header = {'extraFields': slice(3, 8), 'readCount': 8, 'meanCoverage': 9,
              'thresholds': {}, 'sampleName': 10}
    row = BASE_COLS + EXTRA_COLS + SAMBAMBA_COLS + SAMBAMBA_END_COLS
    row_data = sambamba.expand_row(header, row)
    assert row_data['extraFields'] == EXTRA_COLS
    assert row_data['readCount'] == 232
    assert row_data['meanCoverage'] == 25.4946
    assert row_data['sampleName'] == 'ADM992A10'


def test_depth_output_exon():
    bed_lines = ["# chrom\tchromStart\tchromEnd\tF3\tF4\tF5\tF6\tF7\t"
                    "readCount\tmeanCoverage\tpercentage10\tpercentage20"
                    "\tpercentage100\tsampleName",
                 "1\t69089\t70007\t1-69090-70007\t0\t+\tCCDS30547.1\tOR4F5\t"
                    "232\t25.4946\t57.9521\t36.0566\t5.55556\tADM992A10\t"]

    row_data = sambamba.depth_output(bed_lines)
    results = [data for data in row_data]
    assert len(results) == 1
    assert results[0]['sampleName'] == 'ADM992A10'
    assert results[0]['chromStart'] == 69089


def test_depth_output_gene():
    bed_lines = ["# chrom\tchromStart\tchromEnd\tF3\treadCount\tmeanCoverage\t"
                    "percentage10\tpercentage20\tpercentage100\tsampleName",
                 "1\t69089\t70007\tOR4F5\t232\t25.4946\t57.9521\t36.0566\t"
                    "5.55556\tADM992A10\t"]

    row_data = sambamba.depth_output(bed_lines)
    results = [data for data in row_data]
    assert len(results) == 1
    assert results[0]['chrom'] == '1'
    assert results[0]['chromStart'] == 69089
    assert results[0]['chromEnd'] == 70007
    assert results[0]['extraFields'] == ['OR4F5']
    assert results[0]['readCount'] == 232
    assert results[0]['meanCoverage'] == 25.4946
    assert results[0]['thresholds'][20] == 36.0566
    assert results[0]['sampleName'] == 'ADM992A10'



# -*- coding: utf-8 -*-
from .bootstrap import bootstrap as bootstrap_command
from .calculate import calculate as calculate_command
from .database import db as database_command
from .init import init as init_command
from .link import link as link_command
from .load import load as load_command
from .root import root as root_command
from .sambamba import sambamba as sambamba_command
from .sex import sex as sex_command



# -*- coding: utf-8 -*-
from chanjo.store import Exon


def _exon_kwargs(data):
    """Parse out the values to fill in the Exon model fields.

    Args:
        data (dict): parsed sambamba output row

    Returns:
        dict: kwargs prepared for Exon model
    """
    exon_id = data.get('name') or data['extraFields'][0]
    return {'exon_id': exon_id, 'chromosome': data['chrom'],
            'start': data['chromStart'], 'end': data['chromEnd']}


def exon(data):
    """Create exon model from an output row.

    Args:
        data (dict): parsed sambamba output row

    Return:
        Exon: filled in Exon database model
    """
    kwargs = _exon_kwargs(data)
    exon_obj = Exon(**kwargs)
    return exon_obj


def get_or_build_exon(session, data):
    """Fetch exon from database or build a new instance.

    Args:
        session (Session): database session object
        data (dict): parsed sambamba output row
    """
    ex_filters = _exon_kwargs(data)
    exon_obj = session.query(Exon).filter_by(**ex_filters).first()
    if exon_obj is None:
        # no existing exon object, create a new one
        exon_obj = exon(data)

    return exon_obj



# -*- coding: utf-8 -*-
import logging

import click
from sqlalchemy.orm.exc import NoResultFound

from chanjo.store import Store
from chanjo.store.models import ExonStatistic, Sample

logger = logging.getLogger(__name__)


@click.group()
@click.pass_context
def db(context):
    """Interact with the database for maintainance tasks."""
    context.db = Store(uri=context.obj['database'])


@db.command()
@click.option('--reset', is_flag=True, help='tear down existing db')
@click.pass_context
def setup(context, reset):
    """Initialize a new datbase from scratch."""
    if reset:
        logger.info('tearing down existing database')
        context.parent.db.tear_down()
    logger.info('setting up new database')
    context.parent.db.set_up()


@db.command()
@click.argument('sample_id', type=str)
@click.pass_context
def remove(context, sample_id):
    """Remove all traces of a sample from the database."""
    db = context.parent.db
    logger.debug('find sample in database with id: %s', sample_id)
    try:
        sample_obj = db.query(Sample).filter_by(sample_id=sample_id).one()
    except NoResultFound:
        logger.warn('sample (%s) not found in database', sample_id)
        context.abort()
    logger.debug('delete all related exon statistics')
    (db.query(ExonStatistic)
       .filter(ExonStatistic.sample_id == sample_obj.id)
       .delete())
    logger.debug('delete sample from database')
    db.session.delete(sample_obj)
    logger.info('remove sample (%s) from database', sample_id)
    db.save()
